@@ -1,12 +1,11 @@
 # For details on how netCDF4 builds on HDF5:
 # https://docs.unidata.ucar.edu/netcdf-c/current/file_format_specifications.html#netcdf_4_spec
-import os.path
+import os
 import warnings
 import weakref
 from collections import ChainMap, Counter, OrderedDict, defaultdict
 from collections.abc import Mapping
 
-import h5py
 import numpy as np
 from packaging import version
 
@@ -24,12 +23,26 @@ from .utils import (
 )
 
 try:
-    import h5pyd
+    import h5py  # noqa
+except ImportError:
+    no_h5py = True
+else:
+    no_h5py = False
+
+try:
+    import h5pyd  # noqa
 except ImportError:
     no_h5pyd = True
 else:
     no_h5pyd = False
 
+try:
+    import pyfive  # noqa
+except ImportError:
+    no_pyfive = True
+else:
+    no_pyfive = False
+
 
 NOT_A_VARIABLE = b"This is a netCDF dimension but not a netCDF variable."
 
@@ -105,6 +118,107 @@ def _expanded_indexer(key, ndim):
     return key[k1] + res_dims + key[k2]
 
 
+def _parse_backend(path, mode, backend, **kwargs):
+    """Parse the 'backend' keyword to File.__init__.
+
+    Parameters
+    ----------
+    path : path-like
+    mode : str
+    backend : str
+        The backend parameter.
+
+    Returns
+    -------
+    backend: str
+        The backend that is going to be used. If the input backend is
+        None, then a value of the H5NETCDF_BACKEND environment
+        variable is used.
+
+    """
+    is_remote = path.startswith(("http", "hdf5:")) if isinstance(path, str) else False
+    driver = kwargs.get("driver")
+    if backend not in (None, "pyfive", "h5py", "h5pyd"):
+        raise ValueError(
+            f"Unknown backend {backend!r} - valid options are: "
+            "'pyfive', 'h5py', 'h5pyd'"
+        )
+
+    if driver == "h5pyd" and backend not in ["pyfive", "h5py"]:
+        msg = "Specifying driver='h5pyd' is deprecated, please use backend='h5pyd' instead."
+        warnings.warn(msg, DeprecationWarning)
+        backend, driver = "h5pyd", None
+
+    if driver is not None:
+        if backend not in [None, "h5py"]:
+            msg = f"driver={driver!r} only works with 'h5py' backend, but given backend={backend}."
+            raise ValueError(msg)
+        else:
+            backend = "h5py"
+
+    if is_remote and backend is None and driver is None:
+        backend = "h5pyd"
+
+    if backend is None:
+        read_backend = os.environ.get("H5NETCDF_READ_BACKEND", "h5py")
+        write_backend = os.environ.get("H5NETCDF_WRITE_BACKEND", "h5py")
+        backend = read_backend if mode == "r" else write_backend
+
+    no_backend = {
+        "pyfive": no_pyfive,
+        "h5py": no_h5py,
+        "h5pyd": no_h5pyd,
+    }
+
+    if no_backend.get(backend, False):
+        raise ImportError(
+            f"No module named {backend!r}, backend not available. "
+            f"Please install {backend!r} into your Python environment."
+        )
+
+    return backend
+
+
+def _get_track_order(backend):
+    """
+    In h5netcdf version 0.12.0 and earlier, order tracking was disabled in
+    HDF5 file. As this is a requirement for the current netCDF4 standard,
+    it has been enabled without deprecation as of version 0.13.0 (:issue:`128`).
+
+    Datasets created with h5netcdf version 0.12.0 that are opened with
+    newer versions of h5netcdf will continue to disable order tracker.
+
+    If an h5py File object is passed in, closing the h5netcdf wrapper will
+    not close the h5py File. In other cases, closing the h5netcdf File object
+    does close the underlying file.
+
+    """
+    # 2022/01/09
+    # netCDF4 wants the track_order parameter to be true
+    # through this might be getting relaxed in a more recent version of the
+    # standard
+    # https://github.com/Unidata/netcdf-c/issues/2054
+    # https://github.com/h5netcdf/h5netcdf/issues/128
+    # h5py versions less than 3.7.0 had a bug that limited the number of
+    # attributes when track_order was set to true by default.
+    # However, setting track_order to True helps with compatibility
+    # with netcdf4-c and generally, keeping track of how things were added
+    # to the dataset.
+    # https://github.com/h5netcdf/h5netcdf/issues/136#issuecomment-1017457067
+    if backend == "h5py":
+        import h5py
+
+        track_order = version.parse(h5py.__version__) >= version.parse("3.7.0")
+    elif backend == "h5pyd":
+        import h5pyd
+
+        track_order = version.parse(h5pyd.__version__) >= version.parse("0.21.0")
+    else:
+        track_order = None
+
+    return track_order
+
+
 class BaseObject:
     def __init__(self, parent, name):
         self._parent_ref = weakref.ref(parent)
@@ -125,6 +239,10 @@ class BaseObject:
         # subclasses:
         return self._root._h5file[self._h5path]
 
+    @property
+    def _backend(self):
+        return self._root._backend
+
     @property
     def name(self):
         """Return object name."""
@@ -148,10 +266,11 @@ _h5type_mapping = {
 
 def _get_h5usertype_identifier(h5type):
     """Return H5 Type Identifier from given H5 Datatype."""
-    try:
-        # h5py first
+    module = h5type.__module__
+    if module.startswith("h5py.") or module.startswith("pyfive."):
+        # h5py/pyfive first
         h5typeid = h5type.id.get_class()
-    except AttributeError:
+    elif module.startswith("h5pyd."):
         # h5pyd second
         h5typeid = _h5type_mapping[h5type.id.type_json["class"]]
     return h5typeid
@@ -159,12 +278,14 @@ def _get_h5usertype_identifier(h5type):
 
 def _get_h5dstype_identifier(h5type):
     """Return H5 Type Identifier from given H5 Dataset."""
-    try:
-        # h5py first
+    module = h5type.__module__
+    if module.startswith("h5py."):
         h5typeid = h5type.id.get_type().get_class()
-    except AttributeError:
-        # h5pyd second
+    elif module.startswith("h5pyd."):
         h5typeid = _h5type_mapping[h5type.id.type_json["class"]]
+    elif module.startswith("pyfive."):
+        h5typeid = h5type.id.get_type().type_id
+
     return h5typeid
 
 
@@ -502,11 +623,13 @@ class BaseVariable(BaseObject):
         """
         # this is really painful as we have to iterate over all types
         # and check equality
-        usertype = self._parent._get_usertype_dict(self._h5type_identifier)
-        if usertype is not None:
-            for tid in usertype.values():
-                if self._h5datatype == tid._h5datatype:
-                    return tid
+        if self._backend is not None:
+            usertype = self._parent._get_usertype_dict(self._h5type_identifier)
+            if usertype is not None:
+                for tid in usertype.values():
+                    if self._h5datatype == tid._h5datatype:
+                        return tid
+
         return self.dtype
 
     def _get_padding(self, key):
@@ -568,14 +691,19 @@ class BaseVariable(BaseObject):
             # fix boolean indexing for affected versions
             # https://github.com/h5py/h5py/pull/2079
             # https://github.com/h5netcdf/h5netcdf/pull/125/
-            h5py_version = version.parse(h5py.__version__)
-            if version.parse("3.0.0") <= h5py_version < version.parse("3.7.0"):
-                key = _transform_1d_boolean_indexers(key)
+            if self._backend == "h5py":
+                h5py_version = version.parse(self._root._h5py.__version__)
+                if version.parse("3.0.0") <= h5py_version < version.parse("3.7.0"):
+                    key = _transform_1d_boolean_indexers(key)
 
         if getattr(self._root, "decode_vlen_strings", False):
             string_info = self._root._h5py.check_string_dtype(self._h5ds.dtype)
             if string_info and string_info.length is None:
-                return self._h5ds.asstr()[key]
+                if self._backend == "pyfive":
+                    # pyfive backend has already dealt with strings
+                    return self._h5ds[key]
+                else:
+                    return self._h5ds.asstr()[key]
 
         # get padding
         padding = self._get_padding(key)
@@ -689,6 +817,54 @@ class Variable(BaseVariable):
     def shuffle(self):
         return self._h5ds.shuffle
 
+    def filters(self):
+        """Return HDF5 filter parameters dictionary, matching netCDF4's var.filters() API.
+
+        Returns a dictionary with compression flags (zlib, szip, bzip2, blosc, zstd)
+        and complevel, matching the format returned by netCDF4's var.filters() method.
+        Returns None if no filters are present.
+        """
+        # These filters were obtained by opening the same file with netCDF4 and
+        # calling the filter method there
+        filters_dict = {
+            "zlib": False,
+            "szip": False,
+            "bzip2": False,
+            "blosc": False,
+            "zstd": False,
+            "shuffle": self._h5ds.shuffle,
+            "fletcher32": self._h5ds.fletcher32,
+        }
+
+        if self._h5ds.compression == "gzip":
+            filters_dict["zlib"] = True
+            filters_dict["complevel"] = self._h5ds.compression_opts
+
+        h5py_filters = getattr(self._h5ds, "_filters", {})
+        # These don't really get defined in h5py
+        # https://github.com/HDFGroup/hdf5_plugins/blob/master/docs/RegisteredFilterPlugins.md
+        HDF5_FILTER_TO_COMPRESSION = {
+            "1": "zlib",
+            "2": "szip",
+            "307": "bzip2",
+            "32004": "blosc",
+            "32015": "zstd",
+            # They might return the filter as a string, not as a code
+            "zlib": "zlib",
+            "szip": "szip",
+            "bzip2": "bzip2",
+            "blosc": "blosc",
+            "zstd": "zstd",
+        }
+        for filter_id, filter_opts in h5py_filters.items():
+            compression_name = HDF5_FILTER_TO_COMPRESSION.get(filter_id, None)
+            if compression_name is not None:
+                filters_dict[compression_name] = True
+                if filter_opts and len(filter_opts) > 0:
+                    filters_dict["complevel"] = filter_opts[0]
+
+        return filters_dict
+
 
 class _LazyObjectLookup(Mapping):
     def __init__(self, parent, object_cls):
@@ -951,7 +1127,36 @@ class Group(Mapping):
         if self._root._phony_dims_mode is not None:
             phony_dims = Counter()
 
-        for k, v in self._h5group.items():
+        for k in self._h5group:
+            if self._root._backend == "pyfive":
+                # Some backends might have unsupported HDF5
+                # features. Either skip over them, or fail.
+                unsupported = self._root._unsupported_hdf5_features
+                # warnings.filterwarnings("error", message=r"^\w+\s+datatype class not supported\.$", category=UserWarning,)
+                try:
+                    v = self._h5group[k]
+                except NotImplementedError as e:
+                    msg = (
+                        f"{self._root.backend} backend: Skipping unsupported type "
+                        f"of HDF5 variable or dimension {k!r}."
+                    )
+                    if unsupported == "warn":
+                        warnings.warn(msg)
+                    elif unsupported == "error":
+                        msg2 = (
+                            "Consider setting unsupported_hdf5_features='skip' or "
+                            "unsupported_hdf5_features='warn'."
+                        )
+                        e.add_note(" ".join((msg, msg2)))
+                        raise e
+                    continue
+                else:
+                    # probably unsupported DataType
+                    if v is None:
+                        continue
+            else:
+                v = self._h5group[k]
+
             if isinstance(v, self._root._h5py.Group):
                 # add to the groups collection if this is a h5py(d) Group
                 # instance
@@ -1008,8 +1213,9 @@ class Group(Mapping):
 
     @property
     def _track_order(self):
-        if self._root._h5py.__name__ == "h5pyd":
-            return False
+        if self._root._backend == "h5pyd":
+            return self._h5group.track_order
+
         # TODO: make a suggestion to upstream to create a property
         # for files to get if they track the order
         # As of version 3.6.0 this property did not exist
@@ -1393,7 +1599,13 @@ class Group(Mapping):
 
     def _get_usertype(self, h5type):
         """Get usertype from related usertype dict."""
-        h5typeid = _get_h5usertype_identifier(h5type)
+        if self._root._h5py.__name__ == "pyfive":
+            if isinstance(h5type.id, self._root._h5py.h5t.TypeEnumID):
+                h5typeid = 8
+            elif isinstance(h5type.id, self._root._h5py.h5t.TypeCompoundID):
+                h5typeid = 6
+        else:
+            h5typeid = _get_h5usertype_identifier(h5type)
         return self._get_usertype_dict(h5typeid).get(h5type.name.split("/")[-1])
 
     def _get_usertype_dict(self, h5typeid):
@@ -1537,6 +1749,66 @@ class Group(Mapping):
         return cmptype
 
 
+def _open_pyfive(path, mode):
+    import pyfive
+
+    try:
+        h5file = pyfive.File(path, mode)
+    except Exception:
+        raise
+    else:
+        return h5file
+
+
+def _open_h5pyd(path, mode, **kwargs):
+    original_mode = mode
+    if mode != "r":
+        kwargs.setdefault("track_order", _get_track_order("h5pyd"))
+    if mode == "a":
+        mode = "r+"  # probe for existing before creating
+    try:
+        h5file = h5pyd.File(path, mode, **kwargs)
+    except OSError:
+        if original_mode == "a":
+            msg = (
+                "Append mode for h5pyd now probes with 'r+' first and "
+                "only falls back to 'w' if the file is missing.\n"
+                "To silence this warning use 'r+' (open-existing) or 'w' "
+                "(create-new) directly."
+            )
+            warnings.warn(msg, UserWarning, stacklevel=2)
+
+            try:
+                h5file = h5pyd.File(path, "w", **kwargs)
+            except Exception:
+                raise
+            else:
+                return h5file, False
+        raise
+    else:
+        return h5file, (mode != "w")
+
+
+def _open_h5py(path, mode, **kwargs):
+    if mode != "r":
+        kwargs.setdefault("track_order", _get_track_order("h5py"))
+    try:
+        if isinstance(path, str):
+            exists = path.startswith(("http", "s3://")) or (
+                os.path.exists(path) and mode != "w"
+            )
+            h5file = h5py.File(path, mode, **kwargs)
+            return h5file, exists, True
+        elif isinstance(path, h5py.File):
+            # already-open file
+            return path, (mode in {"r", "r+", "a"}), False
+        else:  # file-like object
+            h5file = h5py.File(path, mode, **kwargs)
+            return h5file, (mode in {"r", "r+", "a"}), True
+    except Exception:
+        raise
+
+
 class File(Group):
     def __init__(
         self,
@@ -1545,6 +1817,7 @@ class File(Group):
         format="NETCDF4",
         invalid_netcdf=False,
         phony_dims=None,
+        backend=None,
         **kwargs,
     ):
         """NetCDF4 file constructor.
@@ -1570,6 +1843,16 @@ class File(Group):
         phony_dims: 'sort', 'access'
             See :ref:`phony dims` for more details.
 
+        backend: 'h5py', 'h5pyd', 'pyfive' or None
+            The default backend is h5py (backend=None, or backend='h5py'), but
+            for reading data, the pure python pyfive backend is available.
+
+        unsupported_hdf5_features: str
+            How h5netcdf handles pyfive's unsupported hdf5_features
+            'skip': skip, no warning
+            'warn': skip, with warning
+            'error' raise an exception
+
         track_order: bool
             Corresponds to the h5py.File `track_order` parameter. Unless
             specified, the library will choose a default that enhances
@@ -1579,10 +1862,13 @@ class File(Group):
             append to a file. If an older version of h5py is detected, this
             parameter will be set to False by default to work around a bug in
             h5py limiting the number of attributes for a given variable.
+            Ignored for the 'pyfive' backend.
 
         **kwargs:
-            Additional keyword arguments to be passed to the ``h5py.File``
-            constructor.
+            Additional keyword arguments to be passed to the backend
+            file constructor, which is ``h5py.File`` for the 'h5py'
+            backend (the default), or ``pyfive.File`` for the 'pyfive'
+            backend.
 
         Notes
         -----
@@ -1596,82 +1882,32 @@ class File(Group):
         If an h5py File object is passed in, closing the h5netcdf wrapper will
         not close the h5py File. In other cases, closing the h5netcdf File object
         does close the underlying file.
-        """
-        # 2022/01/09
-        # netCDF4 wants the track_order parameter to be true
-        # through this might be getting relaxed in a more recent version of the
-        # standard
-        # https://github.com/Unidata/netcdf-c/issues/2054
-        # https://github.com/h5netcdf/h5netcdf/issues/128
-        # h5py versions less than 3.7.0 had a bug that limited the number of
-        # attributes when track_order was set to true by default.
-        # However, setting track_order to True helps with compatibility
-        # with netcdf4-c and generally, keeping track of how things were added
-        # to the dataset.
-        # https://github.com/h5netcdf/h5netcdf/issues/136#issuecomment-1017457067
-        track_order_default = version.parse(h5py.__version__) >= version.parse("3.7.0")
-        track_order = kwargs.pop("track_order", track_order_default)
 
+        """
         self.decode_vlen_strings = kwargs.pop("decode_vlen_strings", None)
         self._close_h5file = True
+        self._preexisting_file = True
+
         try:
-            if isinstance(path, str):
-                if kwargs.get("driver") == "h5pyd" or (
-                    path.startswith(("http://", "https://", "hdf5://"))
-                    and "driver" not in kwargs
-                ):
-                    if no_h5pyd:
-                        raise ImportError(
-                            "No module named 'h5pyd'. h5pyd is required for "
-                            f"opening urls: {path}"
-                        )
-                    self._preexisting_file = mode in {"r", "r+", "a"}
-                    # remap "a" -> "r+" to check file existence
-                    # fallback to "w" if not
-                    _mode = mode
-                    if mode == "a":
-                        mode = "r+"
-                    self._h5py = h5pyd
-                    try:
-                        self.__h5file = self._h5py.File(
-                            path, mode, track_order=track_order, **kwargs
-                        )
-                        self._preexisting_file = mode != "w"
-                    except OSError:
-                        # if file does not exist, create it
-                        if _mode == "a":
-                            mode = "w"
-                            self.__h5file = self._h5py.File(
-                                path, mode, track_order=track_order, **kwargs
-                            )
-                            self._preexisting_file = False
-                            msg = (
-                                "Append mode for h5pyd now probes with 'r+' first and "
-                                "only falls back to 'w' if the file is missing.\n"
-                                "To silence this warning use 'r+' (open-existing) or 'w' "
-                                "(create-new) directly."
-                            )
-                            warnings.warn(msg, UserWarning, stacklevel=2)
-                        else:
-                            raise
-                else:
-                    self._preexisting_file = os.path.exists(path) and mode != "w"
-                    self._h5py = h5py
-                    self.__h5file = self._h5py.File(
-                        path, mode, track_order=track_order, **kwargs
-                    )
-            elif isinstance(path, h5py.File):
-                self._preexisting_file = mode in {"r", "r+", "a"}
-                self._h5py = h5py
-                self.__h5file = path
-                # h5py File passed in: let the caller decide when to close it
-                self._close_h5file = False
-            else:  # file-like object
-                self._preexisting_file = mode in {"r", "r+", "a"}
-                self._h5py = h5py
-                self.__h5file = self._h5py.File(
-                    path, mode, track_order=track_order, **kwargs
+            self._backend = _parse_backend(path, mode, backend, **kwargs)
+            if self.backend == "pyfive":
+                self._unsupported_hdf5_features = kwargs.pop(
+                    "unsupported_hdf5_features",
+                    os.environ.get("PYFIVE_UNSUPPORTED_FEATURE", "error"),
                 )
+                self._h5py = pyfive
+                self.__h5file = _open_pyfive(path, mode)
+            elif self.backend == "h5pyd":
+                self._h5py = h5pyd
+                self.__h5file, self._preexisting_file = _open_h5pyd(
+                    path, mode, **kwargs
+                )
+            else:  # default h5py
+                self._h5py = h5py
+                self.__h5file, self._preexisting_file, self._close_h5file = _open_h5py(
+                    path, mode, **kwargs
+                )
+
         except Exception:
             self._closed = True
             raise
@@ -1791,6 +2027,17 @@ class File(Group):
     def _root(self):
         return self
 
+    @property
+    def backend(self) -> str:
+        """The HDF5 backend.
+
+        Returns either "h5py" (the backend is h5py, built on the HDF5
+        C library), "h5pyd" (python library for HDF REST API) or
+        "pyfive" (the backend is pyfive, which is pure Python).
+
+        """
+        return self._backend
+
     def flush(self):
         if self._writable:
             # only write `_NCProperties` in newly created files
@@ -1861,7 +2108,9 @@ class File(Group):
         if self._closed:
             return f"<Closed {self._cls_name}>"
         header = (
-            f"<{self._cls_name} {os.path.basename(self.filename)!r} (mode {self.mode})>"
+            f"<{self._cls_name} "
+            f"{os.path.basename(self.filename)!r} "
+            f"(mode {self.mode}, backend {self.backend})>"
         )
         return "\n".join([header] + self._repr_body())
 
